--[[
Copyright (C) braeven, 2022

Author: braeven
Thanks for Helping: Achimobil, TethysSaturn, DickerSauerlaender
Date: 25.04.2022
Version: 1.0.0.0

Contact/Help/Tutorials:
discord.gg/gHmnFZAypk


Changelog:
1.0.0.0 @ 25.04.2022 - Adjusted to Modhub Guidelines
1.0.0.0 @ 19.04.2022 - Adjusted to 1.4 changes
1.0.0.0 @ 16.04.2022 - Initial Release


Important:.
No changes are allowed to this script without permission from Braeven.
If you want to make a production with this script, look in the revamp.txt/revamp_en.txt in this file or download the FS22_Revamp_Production Packs for reference
Don't copy the script into a production, load the mod as a dependency!

Es dürfen an diesem Script keine Veränderungen ohne Erlaubnis von Braeven gemacht werden.
Wenn du eine Produktion mit diesem Script bauen möchtest, lese dir die revamp.txt durch oder guck dir die FS22_Revamp_Production Packs an.
Nicht das Script in Produktionen kopieren, ladet den Mod über eine Dependency!

]]



Revamp = {};
ProductionPoint.OUTPUT_MODE.STORE = 3



--Production Revamp: Neue XML-Variablen anmelden die aus der Produktion ausgelesen werden sollen
function Revamp.registerXMLPaths(schema, basePath)
  schema:register(XMLValueType.STRING, basePath .. ".productions.production(?).inputs.input(?)#mix", "Mixing Group for Input", 0)
  schema:register(XMLValueType.STRING, basePath .. ".productions.production(?).inputs.input(?)#boostfactor", "Boost-Factor for Input", 0)
  schema:register(XMLValueType.BOOL, basePath .. ".productions.production(?).outputs.output(?)#boost", "Activate Boost for Output", true)
end

ProductionPoint.registerXMLPaths = Utils.prependedFunction(ProductionPoint.registerXMLPaths, Revamp.registerXMLPaths)



--Production Revamp: Neuen Savegame-Variable anlegen, damit der Modus "Einlagern" ausgelesen/gespeichert werden kann
function Revamp.registerSavegameXMLPaths(schema, basePath)
  schema:register(XMLValueType.STRING, basePath .. ".storageFillType(?)", "fillType currently configured to be stored")
end

ProductionPoint.registerSavegameXMLPaths = Utils.prependedFunction(ProductionPoint.registerSavegameXMLPaths, Revamp.registerSavegameXMLPaths)
print("Production Revamp: XML-Paths registered");



--Production Revamp: Load Funktion überschrieben um die neuen Einstellungen aus den Produktionen einlesen zu können und die "Einlagern" Funktion hinzugefügt
function Revamp:load(superFunc, components, xmlFile, key, customEnv, i3dMappings)
  self.node = components[1].node
  local name = xmlFile:getValue(key .. "#name")
  self.name = name and g_i18n:convertText(name, customEnv)
  self.productions = {}
  self.productionsIdToObj = {}
  self.inputFillTypeIds = {}
  self.inputFillTypeIdsArray = {}
  
  self.outputFillTypeIds = {}
  self.outputFillTypeIdsArray = {}
  self.outputFillTypeIdsDirectSell = {}
  self.outputFillTypeIdsAutoDeliver = {}
  self.outputFillTypeIdsToPallets = {}
  --Production Revamp: Storage-Liste hinzugefügt für "Einlagern" Funktion bei Produktionen
  self.outputFillTypeIdsStorage = {}  
  self.sharedThroughputCapacity = xmlFile:getValue(key .. ".productions#sharedThroughputCapacity", true)
  local usedProdIds = {}

  xmlFile:iterate(key .. ".productions.production", function (index, productionKey)
    local production = {
      id = xmlFile:getValue(productionKey .. "#id"),
      name = xmlFile:getValue(productionKey .. "#name", nil, customEnv, false),
    }
    local params = xmlFile:getValue(productionKey .. "#params")

    if params ~= nil then
      params = params:split("|")

      for i = 1, #params do
        params[i] = g_i18n:convertText(params[i], customEnv)
      end

      production.name = string.format(production.name, unpack(params))
    end 

    if not production.id then
      Logging.xmlError(xmlFile, "missing id for production '%s'", production.name or index)

      return false
    end

    for i = 1, #usedProdIds do
      if usedProdIds[i] == production.id then
        Logging.xmlError(xmlFile, "production id '%s' already in use", production.id)

        return false
      end
    end

    table.insert(usedProdIds, production.id)

    local cyclesPerMonth = xmlFile:getValue(productionKey .. "#cyclesPerMonth")
    local cyclesPerHour = xmlFile:getValue(productionKey .. "#cyclesPerHour")
    local cyclesPerMinute = xmlFile:getValue(productionKey .. "#cyclesPerMinute")
    production.cyclesPerMinute = cyclesPerMonth and cyclesPerMonth / 60 / 24 or cyclesPerHour and cyclesPerHour / 60 or cyclesPerMinute or 1
    production.cyclesPerHour = cyclesPerHour or production.cyclesPerMinute * 60
    production.cyclesPerMonth = cyclesPerMonth or production.cyclesPerHour * 24
    local costsPerActiveMinute = xmlFile:getValue(productionKey .. "#costsPerActiveMinute")
    local costsPerActiveHour = xmlFile:getValue(productionKey .. "#costsPerActiveHour")
    local costsPerActiveMonth = xmlFile:getValue(productionKey .. "#costsPerActiveMonth")
    production.costsPerActiveMinute = costsPerActiveMonth and costsPerActiveMonth / 60 / 24 or costsPerActiveHour and costsPerActiveHour / 60 or costsPerActiveMinute or 1
    production.costsPerActiveHour = costsPerActiveHour or production.costsPerActiveMinute * 60
    production.costsPerActiveMonth = costsPerActiveMonth or production.costsPerActiveHour * 24
    production.status = ProductionPoint.PROD_STATUS.INACTIVE
    production.inputs = {}

    xmlFile:iterate(productionKey .. ".inputs.input", function (inputIndex, inputKey)
      local input = {}
      local fillTypeString = xmlFile:getValue(inputKey .. "#fillType")
      input.type = g_fillTypeManager:getFillTypeIndexByName(fillTypeString)

      if input.type == nil then
        Logging.xmlError(xmlFile, "Unable to load fillType '%s' for '%s'", fillTypeString, inputKey)
      else
        --Production Revamp: mix und boostfactor aus der XML auslesen und in der input-Liste hinterlegen
        input.mix = xmlFile:getValue(inputKey .. "#mix", 0)
        if input.mix == "boost" then
          input.mix = 6
        else
          input.mix = tonumber(input.mix)
        end
        input.boostfactor = xmlFile:getValue(inputKey .. "#boostfactor", 0)

        self.inputFillTypeIds[input.type] = true

        table.addElement(self.inputFillTypeIdsArray, input.type)


        input.amount = xmlFile:getValue(inputKey .. "#amount", 1)
        table.insert(production.inputs, input)
      end
    end)

    if #production.inputs == 0 then
      Logging.xmlError(xmlFile, "No inputs for production '%s'", productionKey)

      return
    end

    production.outputs = {}
    production.primaryProductFillType = nil
    local maxOutputAmount = 0

    xmlFile:iterate(productionKey .. ".outputs.output", function (outputIndex, outputKey)
      local output = {}
      local fillTypeString = xmlFile:getValue(outputKey .. "#fillType")
      output.type = g_fillTypeManager:getFillTypeIndexByName(fillTypeString)

      if output.type == nil then
        Logging.xmlError(xmlFile, "Unable to load fillType '%s' for '%s'", fillTypeString, outputKey)
      else
        output.sellDirectly = xmlFile:getValue(outputKey .. "#sellDirectly", false)
        --Production Revamp: boost aus der XML auslesen und in der Output-Liste hinterlegen
        output.boost = xmlFile:getValue(outputKey .. "#boost", true)

        if not output.sellDirectly then
          self.outputFillTypeIds[output.type] = true

          table.addElement(self.outputFillTypeIdsArray, output.type)
        else
          self.soldFillTypesToPayOut[output.type] = 0
        end

        output.amount = xmlFile:getValue(outputKey .. "#amount", 1)

        table.insert(production.outputs, output)

        if maxOutputAmount < output.amount then
          production.primaryProductFillType = output.type
          maxOutputAmount = output.amount
        end
      end
    end)

    if #production.outputs == 0 then
      Logging.xmlError(xmlFile, "No outputs for production '%s'", productionKey)
    end

    if self.isClient then
      production.samples = {
        active = g_soundManager:loadSampleFromXML(xmlFile, productionKey .. ".sounds", "active", self.baseDirectory, components, 1, AudioGroup.ENVIRONMENT, i3dMappings, nil)
      }
      production.animationNodes = g_animationManager:loadAnimations(xmlFile, productionKey .. ".animationNodes", components, self, i3dMappings)
      production.effects = g_effectManager:loadEffect(xmlFile, productionKey .. ".effectNodes", components, self, i3dMappings)
    end

    if self.productionsIdToObj[production.id] ~= nil then
      Logging.xmlError(xmlFile, "Error: production id '%s' already used", production.id)

      return false
    end

    self.productionsIdToObj[production.id] = production

    table.insert(self.productions, production)
    return true
  end)

  if #self.productions == 0 then
    Logging.xmlError(xmlFile, "No valid productions defined")
  end

  if self.owningPlaceable == nil then
    print("Error: ProductionPoint.owningPlaceable was not set before load()")

    return false
  end

  self.interactionTriggerNode = xmlFile:getValue(key .. ".playerTrigger#node", nil, components, i3dMappings)

  if self.interactionTriggerNode ~= nil then
    addTrigger(self.interactionTriggerNode, "interactionTriggerCallback", self)
  end

  if self.isClient then
    self.samples = {
      idle = g_soundManager:loadSampleFromXML(xmlFile, key .. ".sounds", "idle", self.baseDirectory, components, 1, AudioGroup.ENVIRONMENT, i3dMappings, nil),
      active = g_soundManager:loadSampleFromXML(xmlFile, key .. ".sounds", "active", self.baseDirectory, components, 1, AudioGroup.ENVIRONMENT, i3dMappings, nil)
    }
    self.animationNodes = g_animationManager:loadAnimations(xmlFile, key .. ".animationNodes", components, self, i3dMappings)
    self.effects = g_effectManager:loadEffect(xmlFile, key .. ".effectNodes", components, self, i3dMappings)
  end

  self.unloadingStation = SellingStation.new(self.isServer, self.isClient)

  self.unloadingStation:load(components, xmlFile, key .. ".sellingStation", self.customEnvironment, i3dMappings, components[1].node)

  self.unloadingStation.storeSoldGoods = true
  self.unloadingStation.skipSell = self.owningPlaceable:getOwnerFarmId() ~= AccessHandler.EVERYONE

  function self.unloadingStation.getIsFillAllowedFromFarm(_, farmId)
    --Production Revamp: Überschrieben um Missionen und Verkaufen zu ermöglichen
    --return g_currentMission.accessHandler:canFarmAccess(farmId, self.owningPlaceable)
    return true
  end

  self.unloadingStation:register(true)

  local loadingStationKey = key .. ".loadingStation"

  if xmlFile:hasProperty(loadingStationKey) then
    self.loadingStation = LoadingStation.new(self.isServer, self.isClient)

    if not self.loadingStation:load(components, xmlFile, loadingStationKey, self.customEnvironment, i3dMappings, components[1].node) then
      Logging.xmlError(xmlFile, "Unable to load loading station %s", loadingStationKey)

      return false
    end

    function self.loadingStation.hasFarmAccessToStorage(_, farmId)
      return farmId == self.owningPlaceable:getOwnerFarmId()
    end

    self.loadingStation.owningPlaceable = self.owningPlaceable

    self.loadingStation:register(true)
  end

  local palletSpawnerKey = key .. ".palletSpawner"

  if xmlFile:hasProperty(palletSpawnerKey) then
    self.palletSpawner = PalletSpawner.new(self.baseDirectory)

    if not self.palletSpawner:load(components, xmlFile, key .. ".palletSpawner", self.customEnvironment, i3dMappings) then
      Logging.xmlError(xmlFile, "Unable to load pallet spawner %s", palletSpawnerKey)

      return false
    end
  end

  if self.loadingStation == nil and self.palletSpawner == nil then
    Logging.xmlError(xmlFile, "No loading station or pallet spawner for production point")

    return false
  end

  if self.palletSpawner ~= nil then
    for fillTypeId, pallet in pairs(self.palletSpawner:getSupportedFillTypes()) do
      if self.outputFillTypeIds[fillTypeId] then
        self.outputFillTypeIdsToPallets[fillTypeId] = pallet
      end
    end
  end

  self.storage = Storage.new(self.isServer, self.isClient)

  self.storage:load(components, xmlFile, key .. ".storage", i3dMappings)
  self.storage:register(true)

  if self.loadingStation ~= nil then
    if not self.loadingStation:addSourceStorage(self.storage) then
    Logging.xmlWarning(xmlFile, "Unable to add source storage ")
    end
    
    g_currentMission.storageSystem:addLoadingStation(self.loadingStation, self.owningPlaceable)
  end

  self.unloadingStation:addTargetStorage(self.storage)

  for inputFillTypeIndex in pairs(self.inputFillTypeIds) do
    if not self.unloadingStation:getIsFillTypeSupported(inputFillTypeIndex) then
      local fillTypeName = g_fillTypeManager:getFillTypeNameByIndex(inputFillTypeIndex)

      Logging.xmlWarning(xmlFile, "Input filltype '%s' is not supported by unloading station", fillTypeName)
    end
  end

  for outputFillTypeIndex in pairs(self.outputFillTypeIds) do
    if (self.loadingStation == nil or not self.loadingStation:getIsFillTypeSupported(outputFillTypeIndex)) and self.outputFillTypeIdsToPallets[outputFillTypeIndex] == nil then
      local fillTypeName = g_fillTypeManager:getFillTypeNameByIndex(outputFillTypeIndex)

      Logging.xmlWarning(xmlFile, "Output filltype '%s' is not supported by loading station or pallet spawner", fillTypeName)
    end
  end

  self.unloadingStation.owningPlaceable = self.owningPlaceable

  g_currentMission.storageSystem:addUnloadingStation(self.unloadingStation, self.owningPlaceable)
  g_currentMission.economyManager:addSellingStation(self.unloadingStation)

  for i = 1, #self.productions do
    local production = self.productions[i]

    for x = 1, #production.inputs do
      local input = production.inputs[x]

      if not self.storage:getIsFillTypeSupported(input.type) then
        Logging.xmlError(xmlFile, "production point storage does not support fillType '%s' used as in input in production '%s'", g_fillTypeManager:getFillTypeNameByIndex(input.type), production.name)

        return false
      end
    end

    for x = 1, #production.outputs do
      local output = production.outputs[x]

      if not output.sellDirectly and not self.storage:getIsFillTypeSupported(output.type) then
        Logging.xmlError(xmlFile, "production point storage does not support fillType '%s' used as an output in production '%s'", g_fillTypeManager:getFillTypeNameByIndex(output.type), production.name)

        return false
      end
    end
  end

  for supportedFillType, _ in pairs(self.storage:getSupportedFillTypes()) do
    if not self.inputFillTypeIds[supportedFillType] and not self.outputFillTypeIds[supportedFillType] then
      Logging.xmlWarning(xmlFile, "storage fillType '%s' not used as a production input or ouput", g_fillTypeManager:getFillTypeNameByIndex(supportedFillType))
    end
  end

  return true
end

ProductionPoint.load = Utils.overwrittenFunction(ProductionPoint.load, Revamp.load)
print("Production Revamp: Load overwritten");



--Production Revamp: Überschrieben um neues Produktionsscript einfügen zu können
function Revamp:updateProduction()
  if self.lastUpdatedTime == nil then
    self.lastUpdatedTime = g_time
    return
  end

  local dt = MathUtil.clamp(g_time - self.lastUpdatedTime, 0, 30000)
  local timeAdjust = g_currentMission.environment.timeAdjustment
  local numActiveProductions = #self.activeProductions

  if numActiveProductions > 0 then
    local minuteFactorTimescaledDt = dt * self.minuteFactorTimescaled * timeAdjust
    local minuteFactorDt = dt / 60000 * timeAdjust

    for n = 1, numActiveProductions do
      local production = self.activeProductions[n]
      local cyclesPerMinuteMinuteFactor = production.cyclesPerMinute * minuteFactorTimescaledDt
      local cyclesPerMinuteFactorNoTimescale = production.cyclesPerMinute * minuteFactorDt
      --Production Revamp: mix-gruppen/boostgruppe hinzufügen und auf false stellen
      local enoughInputResources = true
      local enoughInputMix1 = false
      local enoughInputMix2 = false
      local enoughInputMix3 = false
      local enoughInputMix4 = false
      local enoughInputMix5 = false
      local enoughInputBoost = false
      local enoughOutputSpace = true
      local maxnum = "0"
      local boost = false
      local booster = "1"

      --Production Revamp: Dieser Abschnitt kontrolliert, ob ein /alle notwendigen Inputs in den jeweiligen Gruppe vorhanden sind
      for x = 1, #production.inputs do
        local input = production.inputs[x]
        local fillLevel = self:getFillLevel(input.type)
        self.inputFillLevels[input] = fillLevel
        local factor = cyclesPerMinuteFactorNoTimescale
        local needamount = input.amount * factor

        if input.mix == 5 then
          maxnum = "5"
          if fillLevel > needamount then
            enoughInputMix5 = true
          end
        elseif input.mix == 4 then
          if maxnum < "5" then
            maxnum = "4"
          end
          if fillLevel > needamount then
            enoughInputMix4 = true
          end
        elseif input.mix == 3 then
          if maxnum < "4" then
            maxnum = "3"
          end
          if fillLevel > needamount then
            enoughInputMix3 = true
          end
        elseif input.mix == 2 then
          if maxnum < "3" then
            maxnum = "2"
          end
          if fillLevel > needamount then
             enoughInputMix2 = true
          end
        elseif input.mix == 1 then
          if maxnum < "2" then
            maxnum = "1"
          end
          if fillLevel > needamount then
            enoughInputMix1 = true
          end
        elseif input.mix == 6 then
          if fillLevel > needamount then
            boost = true
            booster = booster + input.boostfactor
          end
        elseif fillLevel < needamount then
            enoughInputResources = false
        end
      end

      --Production Revamp: dieser Abschnitt wertet die Ergebnisse der einzelnen Input-Gruppen aus
      if enoughInputResources == true then
        if maxnum == "5" then
          if enoughInputMix5 == false or enoughInputMix4 == false or enoughInputMix3 == false or enoughInputMix2 == false or enoughInputMix1 == false then
            enoughInputResources = false
          end
        elseif maxnum == "4" then
          if enoughInputMix4 == false or enoughInputMix3 == false or enoughInputMix2 == false or enoughInputMix1 == false then
            enoughInputResources = false
          end
        elseif maxnum == "3" then
          if enoughInputMix3 == false or enoughInputMix2 == false or enoughInputMix1 == false then
            enoughInputResources = false
          end
        elseif maxnum == "2" then
          if enoughInputMix2 == false or enoughInputMix1 == false then
            enoughInputResources = false
          end      
        elseif maxnum == "1" then
          if enoughInputMix1 == false then
            enoughInputResources = false
          end
        end
      end

      if enoughInputResources == false and self.isOwned then
        if production.status ~= ProductionPoint.PROD_STATUS.MISSING_INPUTS then
          production.status = ProductionPoint.PROD_STATUS.MISSING_INPUTS

          self.owningPlaceable:productionStatusChanged(production, ProductionPoint.PROD_STATUS.MISSING_INPUTS)
          self:setProductionStatus(production.id, production.status)

          break
        end
      end

      if enoughInputResources and self.isOwned then
        for x = 1, #production.outputs do
          local output = production.outputs[x]

          if not output.sellDirectly then
            local freeCapacity = self.storage:getFreeCapacity(output.type)
            
            --Production Revamp: Erweitert um den booster, um zu Kontrollieren ob genug Platz vorhanden wäre
            if freeCapacity < output.amount * cyclesPerMinuteMinuteFactor * booster then
              enoughOutputSpace = false

              if production.status ~= ProductionPoint.PROD_STATUS.NO_OUTPUT_SPACE then
                production.status = ProductionPoint.PROD_STATUS.NO_OUTPUT_SPACE

                self:setProductionStatus(production.id, production.status)
              end

              break
            end
          end
        end
      end

      if self.isOwned then
        self.productionCostsToClaim = self.productionCostsToClaim + production.costsPerActiveMinute * minuteFactorTimescaledDt
      end

      if not self.isOwned or enoughInputResources and enoughOutputSpace then
        local factor = cyclesPerMinuteMinuteFactor / (self.sharedThroughputCapacity and numActiveProductions or 1)

        --Production Revamp: Dieser Abschnitt entfernt die Verarbeiteten Inputs aus dem Silo
        for y = 1, #production.inputs do
          local input = production.inputs[y]
          local fillLevel = self.inputFillLevels[input]

          if fillLevel > input.amount * cyclesPerMinuteFactorNoTimescale then
            if fillLevel and fillLevel > 0 then
              if input.mix == 5 then
                if enoughInputMix5 then
                  if self.loadingStation ~= nil then
          self.loadingStation:removeFillLevel(input.type, input.amount * factor, self.ownerFarmId)
                  else
                    self.storage:setFillLevel(fillLevel - input.amount * factor, input.type)
                  end
                  enoughInputMix5 = false
                end
              elseif input.mix == 4 then
                if enoughInputMix4 then
                  if self.loadingStation ~= nil then
          self.loadingStation:removeFillLevel(input.type, input.amount * factor, self.ownerFarmId)
                  else
                    self.storage:setFillLevel(fillLevel - input.amount * factor, input.type)
          end
                  enoughInputMix4 = false
                end
              elseif input.mix == 3 then
                if enoughInputMix3 then
                  if self.loadingStation ~= nil then
          self.loadingStation:removeFillLevel(input.type, input.amount * factor, self.ownerFarmId)
                  else
                    self.storage:setFillLevel(fillLevel - input.amount * factor, input.type)
                  end
                  enoughInputMix3 = false
                end
              elseif input.mix == 2 then
                if enoughInputMix2 then
                  if self.loadingStation ~= nil then
          self.loadingStation:removeFillLevel(input.type, input.amount * factor, self.ownerFarmId)
                  else
                    self.storage:setFillLevel(fillLevel - input.amount * factor, input.type)
                  end
                  enoughInputMix2 = false
                end
              elseif input.mix == 1 then
                if enoughInputMix1 then
                  if self.loadingStation ~= nil then
          self.loadingStation:removeFillLevel(input.type, input.amount * factor, self.ownerFarmId)
                  else
                    self.storage:setFillLevel(fillLevel - input.amount * factor, input.type)
                  end
                  enoughInputMix1 = false
                end
              elseif input.mix == 6 then
                if boost then
                  if self.loadingStation ~= nil then
          self.loadingStation:removeFillLevel(input.type, input.amount * factor, self.ownerFarmId)
                  else
                    self.storage:setFillLevel(fillLevel - input.amount * factor, input.type)
                  end
                end
              else
                self.storage:setFillLevel(fillLevel - input.amount * factor, input.type)
              end
            end
          end
        end

        --Production Revamp: Output-Güter dem Silo gutschreiben in Abhängigkeit vom booster, wenn boost nicht deaktiviert wurde für den jeweiligen Filltype
        if self.isOwned then
          for y = 1, #production.outputs do
            local output = production.outputs[y]
            if output.sellDirectly then
              if self.isServer then
                if output.boost then
                  self.soldFillTypesToPayOut[output.type] = self.soldFillTypesToPayOut[output.type] + output.amount * factor * booster
                elseif output.boost == "reverse" then
                  local boosterreverse = booster - 1
                  self.soldFillTypesToPayOut[output.type] = self.soldFillTypesToPayOut[output.type] + output.amount * factor * boosterreverse
                else
                  self.soldFillTypesToPayOut[output.type] = self.soldFillTypesToPayOut[output.type] + output.amount * factor
                end
              end
            else
              local fillLevel = self.storage:getFillLevel(output.type)
              if output.boost then
                self.storage:setFillLevel(fillLevel + output.amount * factor * booster, output.type)
              elseif output.boost == "reverse" then
                local boosterreverse = booster - 1
                self.storage:setFillLevel(fillLevel + output.amount * factor * boosterreverse, output.type)
              else
                self.storage:setFillLevel(fillLevel + output.amount * factor, output.type)
              end
            end
          end
        end

        if production.status ~= ProductionPoint.PROD_STATUS.RUNNING then
          production.status = ProductionPoint.PROD_STATUS.RUNNING

          self.owningPlaceable:productionStatusChanged(production, production.status)
          ProductionPointProductionStatusEvent.sendEvent(self, production.id, production.status)
        end

        table.clear(self.inputFillLevels)
      end
    end
  end

  if self.isServer and self.isOwned and self.palletSpawnCooldown < g_time then
    for fillTypeId, pallet in pairs(self.outputFillTypeIdsToPallets) do
      if self.outputFillTypeIdsDirectSell[fillTypeId] == nil and self.outputFillTypeIdsAutoDeliver[fillTypeId] == nil and self.outputFillTypeIdsStorage[fillTypeId] == nil then
        local fillLevel = self.storage:getFillLevel(fillTypeId)

        if fillLevel > 0 and pallet and pallet.capacity <= fillLevel and not self.waitingForPalletToSpawn then
          self.waitingForPalletToSpawn = true

          self.palletSpawner:spawnPallet(self:getOwnerFarmId(), fillTypeId, self.palletSpawnRequestCallback, self)
        end
      end
    end
  end

  self.lastUpdatedTime = g_time
end

ProductionPoint.updateProduction = Utils.overwrittenFunction(ProductionPoint.updateProduction, Revamp.updateProduction)
print("Production Revamp: Update overwritten");



--Production Revamp: Überschrieben um den Output-Mode auf "Einlagern" setzen zu können
function Revamp:setOutputDistributionMode(superFunc, outputFillTypeId, mode, noEventSend)
  if self.outputFillTypeIds[outputFillTypeId] == nil then
    printf("Error: setOutputDistribution(): fillType '%s' is not an output fillType", g_fillTypeManager:getFillTypeNameByIndex(outputFillTypeId))

    return
  end

  mode = tonumber(mode)
  self.outputFillTypeIdsDirectSell[outputFillTypeId] = nil
  self.outputFillTypeIdsAutoDeliver[outputFillTypeId] = nil
  --Production Revamp: Alle Filltypes generell auf nil setzen für die Storage-Liste
  self.outputFillTypeIdsStorage[outputFillTypeId] = nil

  if mode == ProductionPoint.OUTPUT_MODE.DIRECT_SELL then
    self.outputFillTypeIdsDirectSell[outputFillTypeId] = true
  elseif mode == ProductionPoint.OUTPUT_MODE.AUTO_DELIVER then
    self.outputFillTypeIdsAutoDeliver[outputFillTypeId] = true
  --Production Revamp: Sollte ein Output auf "Einlagern" stehen, Filltype in der Liste hinterlegen
  elseif mode == ProductionPoint.OUTPUT_MODE.STORE then
    self.outputFillTypeIdsStorage[outputFillTypeId] = true
  elseif mode ~= ProductionPoint.OUTPUT_MODE.KEEP then
    printf("Error: setOutputDistribution(): Undefined mode '%s'", mode)

    return
  end

  ProductionPointOutputModeEvent.sendEvent(self, outputFillTypeId, mode, noEventSend)
end

ProductionPoint.setOutputDistributionMode = Utils.overwrittenFunction(ProductionPoint.setOutputDistributionMode, Revamp.setOutputDistributionMode)



--Production Revamp: Überschrieben um den Output-Mode auf "Einlagern" auslesen zu können
function Revamp:getOutputDistributionMode(superFunc, outputFillTypeId)
  if self.outputFillTypeIdsDirectSell[outputFillTypeId] ~= nil then
    return ProductionPoint.OUTPUT_MODE.DIRECT_SELL
  elseif self.outputFillTypeIdsAutoDeliver[outputFillTypeId] ~= nil then
    return ProductionPoint.OUTPUT_MODE.AUTO_DELIVER
  --Production Revamp: Hinzugefügt um Outputs auf Einlagern stellen zu können
  elseif self.outputFillTypeIdsStorage[outputFillTypeId] ~= nil then
    return ProductionPoint.OUTPUT_MODE.STORE
  end

  return ProductionPoint.OUTPUT_MODE.KEEP
end

ProductionPoint.getOutputDistributionMode = Utils.overwrittenFunction(ProductionPoint.getOutputDistributionMode, Revamp.getOutputDistributionMode)
print("Production Revamp: Set & Get DistributionMode Overwritten");



--Production Revamp: LoadFromXML überschrieben um die "Einlagern" Option aus dem Savegame laden zu können
function Revamp:loadFromXMLFile(superFunc, xmlFile, key)
  local palletSpawnCooldown = xmlFile:getValue(key .. "#palletSpawnCooldown")

  if palletSpawnCooldown then
    self.palletSpawnCooldown = g_time + palletSpawnCooldown
  end

  self.productionCostsToClaim = xmlFile:getValue(key .. "#productionCostsToClaim") or self.productionCostsToClaim

  if self.owningPlaceable.ownerFarmId == AccessHandler.EVERYONE then
    for n = 1, #self.productions do
      self:setProductionState(self.productions[n].id, true)
    end
  end

  xmlFile:iterate(key .. ".production", function (index, productionKey)
    local prodId = xmlFile:getValue(productionKey .. "#id")
    local isEnabled = xmlFile:getValue(productionKey .. "#isEnabled")

    if self.productionsIdToObj[prodId] == nil then
      Logging.xmlWarning(xmlFile, "Unknown production id '%s'", prodId)
    else
      self:setProductionState(prodId, isEnabled)
    end
  end)
  xmlFile:iterate(key .. ".directSellFillType", function (index, directSellKey)
    local fillType = g_fillTypeManager:getFillTypeIndexByName(xmlFile:getValue(directSellKey))

    if fillType then
      self:setOutputDistributionMode(fillType, ProductionPoint.OUTPUT_MODE.DIRECT_SELL)
    end
  end)
  xmlFile:iterate(key .. ".autoDeliverFillType", function (index, autoDeliverKey)
    local fillType = g_fillTypeManager:getFillTypeIndexByName(xmlFile:getValue(autoDeliverKey))

    if fillType then
      self:setOutputDistributionMode(fillType, ProductionPoint.OUTPUT_MODE.AUTO_DELIVER)
    end
  end)
  --Production Revamp: Hinzugefügt um die "Einlagern" Option aus dem Savegame laden zu können
  xmlFile:iterate(key .. ".storageFillType", function (index, storageKey)
    local fillType = g_fillTypeManager:getFillTypeIndexByName(xmlFile:getValue(storageKey))

    if fillType then
      self:setOutputDistributionMode(fillType, ProductionPoint.OUTPUT_MODE.STORE)
    end
  end)
  if not self.storage:loadFromXMLFile(xmlFile, key .. ".storage") then
    return false
  end

  return true
end

ProductionPoint.loadFromXMLFile = Utils.overwrittenFunction(ProductionPoint.loadFromXMLFile, Revamp.loadFromXMLFile)



--Production Revamp: SaveToXML überschrieben um die Einstellung "Einlagern" speichern zu können
function Revamp:saveToXMLFile(superFunc, xmlFile, key, usedModNames)
  if g_time < self.palletSpawnCooldown then
    xmlFile:setValue(key .. "#palletSpawnCooldown", self.palletSpawnCooldown - g_time)
  end

  if self.productionCostsToClaim ~= 0 then
    xmlFile:setValue(key .. "#productionCostsToClaim", self.productionCostsToClaim)
  end

  local xmlIndex = 0

  for i = 1, #self.activeProductions do
    local production = self.activeProductions[i]
    local productionKey = string.format("%s.production(%i)", key, xmlIndex)

    xmlFile:setValue(productionKey .. "#id", production.id)
    xmlFile:setValue(productionKey .. "#isEnabled", true)

    xmlIndex = xmlIndex + 1
  end

  xmlFile:setTable(key .. ".directSellFillType", self.outputFillTypeIdsDirectSell, function (fillTypeKey, _, fillTypeId)
    local fillType = g_fillTypeManager:getFillTypeNameByIndex(fillTypeId)

    xmlFile:setValue(fillTypeKey, fillType)
  end)
  xmlFile:setTable(key .. ".autoDeliverFillType", self.outputFillTypeIdsAutoDeliver, function (fillTypeKey, _, fillTypeId)
    local fillType = g_fillTypeManager:getFillTypeNameByIndex(fillTypeId)

    xmlFile:setValue(fillTypeKey, fillType)
  end)
  --Production Revamp: Hinzugefügt um die "Einlagern" Option im speichern zu können
  xmlFile:setTable(key .. ".storageFillType", self.outputFillTypeIdsStorage, function (fillTypeKey, _, fillTypeId)
    local fillType = g_fillTypeManager:getFillTypeNameByIndex(fillTypeId)

    xmlFile:setValue(fillTypeKey, fillType)
  end)
  self.storage:saveToXMLFile(xmlFile, key .. ".storage", usedModNames)
end

ProductionPoint.saveToXMLFile = Utils.overwrittenFunction(ProductionPoint.saveToXMLFile, Revamp.saveToXMLFile)
print("Production Revamp: Load and Save XML overwritten");



--Production Revamp: readStream überschreiben um die "Einlagern" Auswahl im Multiplayer synchronisieren zu können
function Revamp:readStream(superFunc, streamId, connection)
  ProductionPoint:superClass().readStream(self, streamId, connection)

  if connection:getIsServer() then
    for i = 1, streamReadUInt8(streamId) do
      self:setOutputDistributionMode(streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS), ProductionPoint.OUTPUT_MODE.DIRECT_SELL)
    end

    --Production Revamp: Neue Option "Einlagern" hinzugefügt
    for i = 1, streamReadUInt8(streamId) do
      self:setOutputDistributionMode(streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS), ProductionPoint.OUTPUT_MODE.STORE)
    end

    for i = 1, streamReadUInt8(streamId) do
      self:setOutputDistributionMode(streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS), ProductionPoint.OUTPUT_MODE.AUTO_DELIVER)
    end

    local unloadingStationId = NetworkUtil.readNodeObjectId(streamId)

    self.unloadingStation:readStream(streamId, connection)
    g_client:finishRegisterObject(self.unloadingStation, unloadingStationId)

    if self.loadingStation ~= nil then
      local loadingStationId = NetworkUtil.readNodeObjectId(streamId)

      self.loadingStation:readStream(streamId, connection)
      g_client:finishRegisterObject(self.loadingStation, loadingStationId)
    end

    local storageId = NetworkUtil.readNodeObjectId(streamId)

    self.storage:readStream(streamId, connection)
    g_client:finishRegisterObject(self.storage, storageId)

    for i = 1, streamReadUInt8(streamId) do
      local productionId = streamReadString(streamId)

      self:setProductionState(productionId, true)
      self:setProductionStatus(productionId, streamReadUIntN(streamId, ProductionPoint.PROD_STATUS_NUM_BITS))
    end
	
    self.palletLimitReached = streamReadBool(streamId)
  end
end

ProductionPoint.readStream = Utils.overwrittenFunction(ProductionPoint.readStream, Revamp.readStream)



--Production Revamp: writeStream überschreiben um die "Einlagern" Auswahl im Multiplayer synchronisieren zu können
function Revamp:writeStream(superFunc, streamId, connection)
  ProductionPoint:superClass().writeStream(self, streamId, connection)

  if not connection:getIsServer() then
    streamWriteUInt8(streamId, table.size(self.outputFillTypeIdsDirectSell))

    for directSellFillTypeId in pairs(self.outputFillTypeIdsDirectSell) do
      streamWriteUIntN(streamId, directSellFillTypeId, FillTypeManager.SEND_NUM_BITS)
    end

    --Production Revamp: Neue Option "Einlagern" hinzugefügt
    streamWriteUInt8(streamId, table.size(self.outputFillTypeIdsStorage))

    for autoStorageFillTypeId in pairs(self.outputFillTypeIdsStorage) do
      streamWriteUIntN(streamId, autoStorageFillTypeId, FillTypeManager.SEND_NUM_BITS)
    end

    streamWriteUInt8(streamId, table.size(self.outputFillTypeIdsAutoDeliver))

    for autoDeliverFillTypeId in pairs(self.outputFillTypeIdsAutoDeliver) do
      streamWriteUIntN(streamId, autoDeliverFillTypeId, FillTypeManager.SEND_NUM_BITS)
    end

    NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(self.unloadingStation))
    self.unloadingStation:writeStream(streamId, connection)
    g_server:registerObjectInStream(connection, self.unloadingStation)

    if self.loadingStation ~= nil then
      NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(self.loadingStation))
      self.loadingStation:writeStream(streamId, connection)
      g_server:registerObjectInStream(connection, self.loadingStation)
    end

    NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(self.storage))
    self.storage:writeStream(streamId, connection)
    g_server:registerObjectInStream(connection, self.storage)
    streamWriteUInt8(streamId, #self.activeProductions)

    for i = 1, #self.activeProductions do
      local production = self.activeProductions[i]

      streamWriteString(streamId, production.id)
      streamWriteUIntN(streamId, production.status, ProductionPoint.PROD_STATUS_NUM_BITS)
    end
	
    streamWriteBool(streamId, self.palletLimitReached)
  end
end

ProductionPoint.writeStream = Utils.overwrittenFunction(ProductionPoint.writeStream, Revamp.writeStream)
print("Production Revamp: Gui Overwritten #1")



--Production Revamp: updateDetails überschrieben um neue Darstellung der Rezepte zu ermöglichen
function Revamp:updateDetails(superFunc)
  local production = self:getSelectedProduction()
  local status = production.status
  local statusKey = ProductionPoint.PROD_STATUS_TO_L10N[production.status] or "unknown"
  local statusProfile = "ingameMenuProductionDetailValue"

  if status == ProductionPoint.PROD_STATUS.MISSING_INPUTS then
    statusProfile = "ingameMenuProductionDetailValueError"
  elseif status == ProductionPoint.PROD_STATUS.NO_OUTPUT_SPACE then
    statusProfile = "ingameMenuProductionDetailValueError"
  end

  self.detailProductionStatus:applyProfile(statusProfile)
  self.detailProductionStatus:setLocaKey(statusKey)
  self.detailCyclesPerMonth:setText(MathUtil.round(production.cyclesPerMonth, 2))
  self.detailCostsPerMonth:setValue(production.costsPerActiveMonth)

  local function addIcons(list, layout)
    for i = 1, #layout.elements do
      layout.elements[1]:delete()
    end

    for index, item in ipairs(list) do
      if index > 1 then
        self.recipePlus:clone(layout)
      end
      local count = self.recipeText:clone(layout)
      count:setText(g_i18n:formatNumber(item.amount, 2))

      local fillType = g_fillTypeManager:getFillTypeByIndex(item.type)
      local icon = self.recipeFillIcon:clone(layout)

      icon:setImageFilename(fillType.hudOverlayFilename)
    end

    layout:invalidateLayout()
  end

  --Production Revamp: AddIcons-Funktion für boost- und mix-Inputs und Outputs
  local function myAddIcons(list, layout, mixIndex)
    for i = 1, #layout.elements do
      layout.elements[1]:delete()
    end

    for index, item in ipairs(list) do
      if index > 1 then
        if mixIndex == 1 then
           self.recipePlus:clone(layout)
        else
          local count = self.recipeText:clone(layout)
          count:setText(" || ")
        end
      end

      local count = self.recipeText:clone(layout)
      count:setText(g_i18n:formatNumber(item.amount, 2))

      local fillType = g_fillTypeManager:getFillTypeByIndex(item.type)
      local icon = self.recipeFillIcon:clone(layout)
            
      if mixIndex == 7 then
        local count = self.recipeText:clone(layout)
        count:setText( "(+ " .. g_i18n:formatNumber(item.boostfactor*100, 2) .. "%) ")
      end

      icon:setImageFilename(fillType.hudOverlayFilename)
    end
  end

    --Production Revamp: detailsBox resetten
  for i = #self.detailsBox.elements, 1, -1 do
    if i >= 12 then
      if self.detailsBox.elements[i] ~= nil then
        self.detailsBox.elements[i]:delete()
     end
    end
  end

  --Production Revamp: ProduktionsInputs auslesen und in sortierte Listen speichern
  local myList = {}
  myList[1] = {}
  myList[2] = {}
  myList[3] = {}
  myList[4] = {}
  myList[5] = {}
  myList[6] = {}
  myList[7] = {}
  local item = {}
  for index, item in ipairs(production.inputs) do
    local insert = item.mix +1
    local myListItem = myList[insert];
    table.insert(myListItem, item);
  end


  --Production Revamp: Rezept-Übersicht erstellen
  local lineCounter = 0
  for mixIndex, list in pairs(myList) do
    if list ~= nil and #list ~= 0 then 

      --Production Revamp: PlusZeichen zwischen zwei Zutaten-Listen
      if lineCounter ~= 0 then
        local myElement = FlowLayoutElement.new();
        myElement:copyAttributes(self.detailRecipeInputLayout);
        myElement.absPosition[2] = self.detailRecipeInputLayout.absPosition[2] - (self.detailRecipeInputLayout.absSize[2] * (lineCounter));

        lineCounter = lineCounter + 1;
        self.recipePlus:clone(myElement)
        self.detailsBox:addElement(myElement);
        myElement:invalidateLayout()
      end
        
      local myElement = FlowLayoutElement.new();
      myElement:copyAttributes(self.detailRecipeInputLayout);
      myElement.absPosition[2] = self.detailRecipeInputLayout.absPosition[2] - (self.detailRecipeInputLayout.absSize[2] * (lineCounter));

      lineCounter = lineCounter + 1;
      myAddIcons(list, myElement, mixIndex)
      self.detailsBox:addElement(myElement);

      myElement:invalidateLayout()
    end
  end      

 --Production Revamp: Ergebniss-Pfeil verschieben
 self.detailsBox.elements[10].absPosition[2] = self.detailRecipeInputLayout.absPosition[2] - (self.detailRecipeInputLayout.absSize[2] * (lineCounter));

  --Production Revamp: Ergebniss-Liste verschieben
  self.detailRecipeOutputLayout.absPosition[2] = self.detailRecipeInputLayout.absPosition[2] - (self.detailRecipeInputLayout.absSize[2] * (lineCounter + 1));

  --Production Revamp: Ergebniss-Liste befüllen
  addIcons(production.outputs, self.detailRecipeOutputLayout, 1)


  self.storageList:reloadData()
end

InGameMenuProductionFrame.updateDetails = Utils.overwrittenFunction(InGameMenuProductionFrame.updateDetails, Revamp.updateDetails)
print("Production Revamp: Gui Overwritten #2")



function Revamp:populateCellForItemInSection(superFunc, list, section, index, cell)
  if list == self.productionList then
    local productionPoint = self:getProductionPoints()[section]
    local production = productionPoint.productions[index]
    local fillTypeDesc = g_fillTypeManager:getFillTypeByIndex(production.primaryProductFillType)

    if fillTypeDesc ~= nil then
      cell:getAttribute("icon"):setImageFilename(fillTypeDesc.hudOverlayFilename)
    end

    cell:getAttribute("icon"):setVisible(fillTypeDesc ~= nil)
    cell:getAttribute("name"):setText(production.name or fillTypeDesc.title)

    local status = production.status
    local activityElement = cell:getAttribute("activity")

    if status == ProductionPoint.PROD_STATUS.RUNNING then
      activityElement:applyProfile("ingameMenuProductionProductionActivityActive")
    elseif status == ProductionPoint.PROD_STATUS.MISSING_INPUTS or status == ProductionPoint.PROD_STATUS.NO_OUTPUT_SPACE then
      activityElement:applyProfile("ingameMenuProductionProductionActivityIssue")
    else
      activityElement:applyProfile("ingameMenuProductionProductionActivity")
    end
  else
    local _, productionPoint = self:getSelectedProduction()
    local fillType, isInput = nil

    if section == 1 then
      fillType = self.selectedProductionPoint.inputFillTypeIdsArray[index]
      isInput = true
    else
      fillType = self.selectedProductionPoint.outputFillTypeIdsArray[index]
      isInput = false
    end

    if fillType ~= FillType.UNKNOWN then
      local fillLevel = self.selectedProductionPoint:getFillLevel(fillType)
      local capacity = self.selectedProductionPoint:getCapacity(fillType)
      local fillTypeDesc = g_fillTypeManager:getFillTypeByIndex(fillType)

      cell:getAttribute("icon"):setImageFilename(fillTypeDesc.hudOverlayFilename)
      cell:getAttribute("fillType"):setText(fillTypeDesc.title)
      cell:getAttribute("fillLevel"):setText(self.i18n:formatVolume(fillLevel, 0))

      if not isInput then
        local outputMode = productionPoint:getOutputDistributionMode(fillType)
        local outputModeText = g_i18n:getText("Revamp_Spawn");

        if outputMode == ProductionPoint.OUTPUT_MODE.DIRECT_SELL then
          outputModeText = self.i18n:getText("ui_production_output_selling")
        elseif outputMode == ProductionPoint.OUTPUT_MODE.AUTO_DELIVER then
          outputModeText = self.i18n:getText("ui_production_output_distributing")
        --Production Revamp: Hinzugefügt um die "Einlagern" Option anzeigen zu können
        elseif outputMode == ProductionPoint.OUTPUT_MODE.STORE then
          outputModeText = g_i18n:getText("Revamp_Store");
        end

        cell:getAttribute("outputMode"):setText(outputModeText)
      end

      self:setStatusBarValue(cell:getAttribute("bar"), fillLevel / capacity, isInput)
    end
  end
end

InGameMenuProductionFrame.populateCellForItemInSection = Utils.overwrittenFunction(InGameMenuProductionFrame.populateCellForItemInSection, Revamp.populateCellForItemInSection)
print("Production Revamp: Gui Overwritten #3")



--Production Revamp: II oder Y hinzufügen sollte sharedThroughputCapacity aktiv sein
function Revamp:getName(superFunc, revamp)
  local name = nil
    if self.name ~= nil then
    name = self.name
  else
    name = self.owningPlaceable:getName()
  end
  if revamp == nil then
  
  else
    if self.sharedThroughputCapacity == true then
      name = name .. "    -  Y"
    else
      name = name .. "    -  II"
    end
  end
  return name
end

ProductionPoint.getName = Utils.overwrittenFunction(ProductionPoint.getName, Revamp.getName)


--Production Revamp: Nötig um sharedThroughputCapacity anzeigen zu können
function Revamp:getTitleForSectionHeader(superFunc, list, section)
  if list == self.productionList then
    local productionPoint = self:getProductionPoints()[section]

    return productionPoint:getName(true)
  elseif section == 1 then
    return g_i18n:getText("ui_productions_incomingMaterials")
  else
    return g_i18n:getText("ui_productions_outgoingProducts")
  end
end

InGameMenuProductionFrame.getTitleForSectionHeader = Utils.overwrittenFunction(InGameMenuProductionFrame.getTitleForSectionHeader, Revamp.getTitleForSectionHeader)
print("Production Revamp: Gui Overwritten #4")



--Production Revamp: addFillLevel neu geschrieben, um Produktionen als Verkaufspunkte zu erhalten/Missionsgüter nicht im Lager zu haben
function Revamp:addFillLevelFromTool(superFunc, farmId, deltaFillLevel, fillType, fillInfo, toolType, extraAttributes)
  local movedFillLevel = 0
  if deltaFillLevel > 0 then
    --Mission Abfragen
    local usedByMission = false
    for _, mission in pairs(self.missions) do
      if mission.fillSold ~= nil and mission.fillType == fillType and mission.farmId == farmId then
        mission:fillSold(deltaFillLevel)
        usedByMission = true

        break
      end
    end

    --Überprüfen ob Filltype im Lager angenommen wird
    if self:getIsFillTypeAllowed(fillType, extraAttributes) then
      --Sollte eine Mission für die Produktion aktiv sein
      if usedByMission then
        movedFillLevel = deltaFillLevel
      else
        --Überprüfen ob es sich um eine Produktion handelt
        if self.storeSoldGoods then
          --Besitzer-ID der Produktion ermitteln & Lohnerstatus auslesen
          local ownerFarmId = self.owningPlaceable:getOwnerFarmId()
          local testLohn = AccessHandler:canFarmAccessOtherId(farmId, ownerFarmId)

          --Sollte die Porduktion dem Ablader gehören
          if ownerFarmId == farmId then
            movedFillLevel = SellingStation:superClass().addFillLevelFromTool(self, farmId, deltaFillLevel, fillType, fillInfo, toolType, extraAttributes)
          --Sollte die Produktionen jemand anderes gehören oder niemanden
          elseif ownerFarmId ~= farmId and ownerFarmId ~= 0 then
           --sollte Lohnerstatus vorhanden sein
            if testLohn then
              movedFillLevel = SellingStation:superClass().addFillLevelFromTool(self, farmId, deltaFillLevel, fillType, fillInfo, toolType, extraAttributes)
            else
              movedFillLevel = deltaFillLevel
              self:sellFillType(farmId, movedFillLevel, fillType, toolType, extraAttributes)    
            end
          else
            movedFillLevel = deltaFillLevel
            self:sellFillType(farmId, movedFillLevel, fillType, toolType, extraAttributes)
          end
        else
          movedFillLevel = deltaFillLevel
          self:sellFillType(farmId, movedFillLevel, fillType, toolType, extraAttributes)
          self:startFx(fillType)
        end
      end
    end
  end

  return movedFillLevel
end

SellingStation.addFillLevelFromTool = Utils.overwrittenFunction(SellingStation.addFillLevelFromTool, Revamp.addFillLevelFromTool)